
#define COND_NOT_EQUAL     1
#define COND_EQUAL         2
#define COND_LESS_THAN     3
#define COND_LESS_EQUAL    4
#define COND_GREATER_THAN  5
#define COND_GREATER_EQUAL 6

static struct FieldDef ConditionTable[] = {
   { "<>", COND_NOT_EQUAL },
   { "!=", COND_NOT_EQUAL },
   { "=",  COND_EQUAL },
   { "==", COND_EQUAL },
   { "<",  COND_LESS_THAN },
   { "<=", COND_LESS_EQUAL },
   { ">",  COND_GREATER_THAN },
   { ">=", COND_GREATER_EQUAL },
   { NULL, NULL }
};

static BYTE glSatisfied = FALSE;

/*****************************************************************************
** Internal: TranslateValue()
**
** Translates arguments within the available buffer space of the string.
*/

static void TranslateValue(struct Menu *Self, STRING Source, STRING Buffer, LONG BufferSize)
{
   STRING arg, str;
   LONG pos, len, index, count, i, j;
   UBYTE compare[60];

   if ((Source) AND (Source != Buffer)) StrCopy(Source, Buffer, BufferSize);

   if (Self->LocalArgs) {
      /* Search for an argument.  If there are no arguments
      ** in the string, return immediately.
      */

      for (pos=0; (Buffer[pos]) AND (Buffer[pos] != '{'); pos++);

      if (Buffer[pos] IS '{') {
         count = 0;
         for (i=pos; (Buffer[i]); i++) {
            if (Buffer[i] IS '{') {
               /*** Check against user arguments ***/

               for (index=0; index < Self->TotalArgs; index++) {
                  if ((arg = Self->LocalArgs[index])) {

                     str = Buffer + i + 1;
                     for (j=0; (j < sizeof(compare)) AND (*str) AND (*str != '}'); j++) compare[j] = *str++;
                     compare[j] = 0;

                     if (StrCompare(arg, compare, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        for (len=0; arg[len] != 0; len++);
                        count = 0;

                        StrInsert(arg + len + 1, Buffer, i, j+2, BufferSize);
                        break;
                     }
                  }
               }

               if (Buffer[i] != '{') continue;

               if (++count > 60) {
                  DPrintF("@Translate()","[Menu] Recursion in line %.20s", Buffer);
                  break;
               }
            }
         }
      }
   }

   StrTranslate(Buffer, BufferSize, NULL);
}

/*****************************************************************************
** Internal: IfStatementSatisfied()
*/

static LONG IfStatementSatisfied(struct Menu *Self, struct XMLTag *Tag)
{
   STRING Test, Compare;
   LONG Condition, count, flags, i, j;
   UBYTE save, reverse;
   WORD conditionpos, index;
   struct File *file;
   UBYTE buffer[600];
   OBJECTID object_id;

   reverse = FALSE;
   glSatisfied = FALSE; /* Reset the satisfied variable */
   index = 1;

   if (StrCompare("not", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      reverse = TRUE;
      index++;
   }

   if (StrCompare("exists", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));
      if (*buffer) {
         if (FastFindObject(buffer, NULL, &object_id, 1, &count) IS ERR_Okay) {
            glSatisfied = TRUE;
         }
      }
   }
   else if (StrCompare("fileexists", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));

      if (buffer[0] IS '~') {
         flags = FL_APPROXIMATE;
         i = 1;
      }
      else {
         flags = NULL;
         i = 0;
      }

      if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING, buffer + i,
            FID_Flags|TLONG,      flags,
            TAGEND) IS ERR_Okay) {
         glSatisfied = TRUE;
         Action(AC_Free, file, NULL);
      }
   }
   else if ((StrCompare("directory", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) OR
            (StrCompare("isdirectory", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
      /*** This option checks if a location explicitly refers to a directory ***/
      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));

      if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING, buffer,
            TAGEND) IS ERR_Okay) {
         if (file->Flags & FL_DIRECTORY) glSatisfied = TRUE;
         Action(AC_Free, file, NULL);
      }
   }
   else if (StrCompare("isnull", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));

      if ((!buffer[0]) OR (buffer[0] IS '0')) glSatisfied = TRUE;
      else glSatisfied = FALSE;
   }
   else if (StrCompare("notnull", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));

      if ((!buffer[0]) OR (buffer[0] IS '0')) glSatisfied = FALSE;
      else glSatisfied = TRUE;
   }
   else if (StrCompare("statement", Tag->Attributes[index].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      Condition = NULL;
      Compare   = NULL;
      Test      = NULL;

      /*** Statement field ***/

      TranslateValue(Self, Tag->Attributes[index].Value, buffer, sizeof(buffer));

      /* If the statement starts with a calculation (=), do
      ** not confuse it with a condition character.
      */

      if (buffer[0] IS '=') i = 1;
      else i = 0;

      /*** Find the condition statement ***/

      while (buffer[i]) {
         if ((buffer[i] IS '!') AND (buffer[i+1] IS '=')) break;
         if (buffer[i] IS '>') break;
         if (buffer[i] IS '<') break;
         if (buffer[i] IS '=') break;
         i++;
      }

      /*** Continue if a condition was found ***/

      if (buffer[i]) {
         conditionpos = i;

         /*** Test field ***/

         while ((i > 0) AND (buffer[i-1] IS ' ')) i--;
         save = buffer[i];
         buffer[i] = 0;
         if (*buffer IS '=') {
            UBYTE tmp[1024];
            StrCalculate(buffer+1, NULL, tmp, sizeof(tmp));
            Test = StrClone(tmp);
         }
         else Test = StrClone(buffer);
         buffer[i] = save;

         /*** Condition field ***/

         for (i=conditionpos; (buffer[i] IS '!') OR (buffer[i] IS '=') OR (buffer[i] IS '>') OR (buffer[i] IS '<'); i++);

         save = buffer[i];
         buffer[i] = 0;
         for (j=0; ConditionTable[j].Name; j++) {
            if (StrCompare(buffer+conditionpos, ConditionTable[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               Condition = ConditionTable[j].Value;
               break;
            }
         }
         buffer[i] = save;

         /*** Compare field ***/

         while ((buffer[i]) AND (buffer[i] <= 0x20)) i++; /* skip whitespace */

         if (buffer[i]) Compare = StrClone(buffer+i);

         if ((Test) AND (Compare) AND (Condition)) {
            LONG CompareType, TestType;
            DOUBLE CompareFloat, TestFloat;

            /*** Convert the Compare variable to its specified type ***/

            CompareType = StrDatatype(Compare);
            TestType    = StrDatatype(Test);

            /*** Do the comparison ***/

            if (((TestType IS STR_NUMBER) OR (TestType IS STR_FLOAT)) AND ((CompareType IS STR_NUMBER) OR (CompareType IS STR_FLOAT))) {
               CompareFloat = StrToFloat(Compare);
               TestFloat    = StrToFloat(Test);
               switch(Condition) {
                  case COND_NOT_EQUAL:     if (TestFloat != CompareFloat) glSatisfied = TRUE; break;
                  case COND_EQUAL:         if (TestFloat IS CompareFloat) glSatisfied = TRUE; break;
                  case COND_LESS_THAN:     if (TestFloat <  CompareFloat) glSatisfied = TRUE; break;
                  case COND_LESS_EQUAL:    if (TestFloat <= CompareFloat) glSatisfied = TRUE; break;
                  case COND_GREATER_THAN:  if (TestFloat >  CompareFloat) glSatisfied = TRUE; break;
                  case COND_GREATER_EQUAL: if (TestFloat >= CompareFloat) glSatisfied = TRUE; break;
                  default: DPrintF("@If:","Unsupported condition type %d.", Condition);
               }
            }
            else {
               if (Condition IS COND_EQUAL) {
                  if (StrCompare(Test, Compare, NULL, STR_MATCHLENGTH) IS ERR_Okay) glSatisfied = TRUE;
               }
               else if (Condition IS COND_NOT_EQUAL) {
                  if (StrCompare(Test, Compare, NULL, STR_MATCHLENGTH) != ERR_Okay) glSatisfied = TRUE;
               }
               else DPrintF("@If:","String comparison for condition %d not possible.", Condition);
            }
         }
         else DPrintF("@If:","[Menu] Missing if definitions in statement.");

         if (Test)    FreeMemory(Test);
         if (Compare) FreeMemory(Compare);
      }
   }

   if (reverse) glSatisfied ^= 1;

   return(glSatisfied);
}

/*****************************************************************************
** Internal: ParaseXMLTag()
**
** MenuBase, MenuItem and MenuBreak XML tags can be used to specify the look
** and feel of the menu.
**
** Menu, Item and Break XML tags can be used to specify elements to be placed
** within the menu.
*/

static void ParseXMLTag(struct Menu *Self, OBJECTPTR XML, struct XMLTag *Tag)
{
   struct mtGetXMLString getxml;

   if (StrCompare("if", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (IfStatementSatisfied(Self, Tag)) {
         Tag = Tag->Child;
         while (Tag) {
            ParseXMLTag(Self, XML, Tag);
            Tag = Tag->Next;
         }
      }
   }
   else if (StrCompare("else", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      /*** Execute the contents of the <else> tag if the last <if> statement was not satisfied ***/
      if (glSatisfied IS FALSE) {
         Tag = Tag->Child;
         while (Tag) {
            ParseXMLTag(Self, XML, Tag);
            Tag = Tag->Next;
         }
      }
   }
   else if (StrCompare("menubase", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (Tag->Child) {
         if (Self->prvBaseScript) {
            FreeMemory(Self->prvBaseScript);
            Self->prvBaseScript = NULL;
         }
         getxml.Index = Tag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         Self->prvBaseScript = getxml.Result;
      }
   }
   else if (StrCompare("menudef", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      UBYTE value[500];
      for (Tag=Tag->Child; Tag; Tag=Tag->Next) {
         if (Tag->TotalAttributes > 1) {
            TranslateValue(Self, Tag->Attributes[1].Value, value, sizeof(value));
            SetFieldVariable(Self, Tag->Attributes[0].Name, value);
         }
      }
   }
   else if (StrCompare("menuitem", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (Tag->Child) {
         if (Self->prvItemScript) {
            FreeMemory(Self->prvItemScript);
            Self->prvItemScript = NULL;
         }
         getxml.Index = Tag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         Self->prvItemScript = getxml.Result;
      }
   }
   else if (StrCompare("menubreak", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (Tag->Child) {
         if (Self->prvBreakScript) {
            FreeMemory(Self->prvBreakScript);
            Self->prvBreakScript = NULL;
         }
         getxml.Index = Tag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         Self->prvBreakScript = getxml.Result;
      }
   }
   else if (StrCompare("menu", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      AddXMLItem(Self, XML, Tag);
   }
   else if (StrCompare("item", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      AddXMLItem(Self, XML, Tag);
   }
   else if (StrCompare("break", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      AddXMLItem(Self, XML, Tag);
   }
   else if (StrCompare("translation", Tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      UBYTE filename[300];
      LONG j, i;

      for (j=1; j < Tag->TotalAttributes; j++) {
      	if (StrCompare("language", Tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            StrCopy(Tag->Attributes[j].Value, Self->Language, sizeof(Self->Language));
         }
         else if (StrCompare("dir", Tag->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            StrCopy(Tag->Attributes[j].Value, Self->LanguageDir, sizeof(Self->LanguageDir));
         }
      }

      /* Load a translation file if the menu language does not match the user's
      ** language.
      **
      ** The format is: {scriptdir}/lang/{filename}.languagecode
      **
      ** The translation arguments will be expected to be in the MENU
      ** section of the config file.
      */

      if (StrCompare(glLocale, Self->Language, 0, STR_MATCHLENGTH) != ERR_Okay) {
         for (i=0; Self->Location[i]; i++);
         while ((i > 0) AND (Self->Location[i-1] != ':') AND (Self->Location[i-1] != '/') AND (Self->Location[i-1] != '\\')) i--;

         j = StrCopy(Self->Location, filename, i); /* script dir */
         StrFormat(filename+j, sizeof(filename)-j, "%s/%s.%s", Self->LanguageDir, Self->Location + i, glLocale);

         CreateObject(ID_CONFIG, NF_CHILD, &Self->Translation, NULL,
            FID_Location|TSTRING, filename,
            TAGEND);
      }
   }
}

/*****************************************************************************
** Internal: Translate()
**
** Scans the translation object to see if there is a translation for the text.
*/

static STRING Translate(struct Menu *Self, STRING Text)
{
   struct ConfigEntry *entries;
   LONG amtentries;
   WORD i;

   if (Self->Translation) {
      if ((GetFields(Self->Translation, FID_Entries|TPTR,     &entries,
                                        FID_AmtEntries|TLONG, &amtentries,
                                        TAGEND) IS ERR_Okay) AND (entries)) {
         for (i=0; i < amtentries; i++) {
            if (StrCompare("MENU", entries[i].Section, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               if (StrCompare(Text, entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  return(StrClone(entries[i].Data));
               }
            }
         }
      }
   }

   return(StrClone(Text));
}

/*****************************************************************************
** Internal: DrawMenu()
*/

static void DrawMenu(struct Menu *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Picture *picture;
   struct Font *font;
   LONG x, y;
   WORD i, itemheight;

   font = Self->Font;

   SetField(font, FID_Bitmap, FT_POINTER, Bitmap);

   x = Self->LeftMargin;
   y = Self->TopMargin;

   itemheight = GetItemHeight(Self);

   for (i=0; i < Self->TotalItems; i++) {

      if (Self->Highlight IS i) {
         /*** Draw highlighting rectangle in the background ***/

         ActionTags(MT_DrawRGBRectangle, Bitmap, Self->HighlightLM, Self->Items[i]->YCoord,
            Drawable->Width - Self->HighlightRM - Self->HighlightLM,
            Self->Items[i]->Height-1, Self->prvHighlight.Red,
            Self->prvHighlight.Green, Self->prvHighlight.Blue, TRUE);

         if (Self->Flags & MNF_HIGHLIGHTBORDER) {
            ActionTags(MT_DrawRGBRectangle, Bitmap, Self->HighlightLM, Self->Items[i]->YCoord,
               Drawable->Width - Self->HighlightRM - Self->HighlightLM,
               Self->Items[i]->Height-1, Self->prvHighlightBorder.Red,
               Self->prvHighlightBorder.Green, Self->prvHighlightBorder.Blue, FALSE);
         }
      }
      else if (Self->Items[i]->Flags & MENUITEM_BACKGROUND) {
         /*** If the menu has a background colour, draw it ***/

         ActionTags(MT_DrawRGBRectangle, Bitmap, Self->HighlightLM, Self->Items[i]->YCoord,
            Drawable->Width- Self->HighlightRM - Self->HighlightLM,
            Self->Items[i]->Height, Self->Items[i]->Background.Red,
            Self->Items[i]->Background.Green, Self->Items[i]->Background.Blue, TRUE);
      }

      /*** Draw associated menu icon ***/

      if ((Self->Items[i]->PictureID) AND (Self->Flags & MNF_SHOWIMAGES)) {
         if (AccessObject(Self->Items[i]->PictureID, 4000, (OBJECTPTR *)&picture) IS ERR_Okay) {
            picture->Bitmap->Flags |= BMF_ALPHABLEND;
            ActionTags(MT_BlitArea, picture->Bitmap, Bitmap, NULL, 0, 0,
               picture->Bitmap->Width, picture->Bitmap->Height, x + ((Self->ImageSize - picture->Bitmap->Width)/2), y + ((Self->Items[i]->Height - picture->Bitmap->Height)/2));
            ReleaseObject(picture);
         }
      }

      /*** Set the correct font colour ***/

      if (Self->Highlight IS i) {
         font->ColourRGB->Red   = Self->prvFontHighlight.Red;
         font->ColourRGB->Green = Self->prvFontHighlight.Green;
         font->ColourRGB->Blue  = Self->prvFontHighlight.Blue;
      }
      else if (Self->Items[i]->Flags & MENUITEM_FONTCOLOUR) {
         font->ColourRGB->Red   = Self->Items[i]->Colour.Red;
         font->ColourRGB->Green = Self->Items[i]->Colour.Green;
         font->ColourRGB->Blue  = Self->Items[i]->Colour.Blue;
      }
      else {
         font->ColourRGB->Red   = Self->prvFontColour.Red;
         font->ColourRGB->Green = Self->prvFontColour.Green;
         font->ColourRGB->Blue  = Self->prvFontColour.Blue;
      }

      /*** Draw control key text ***/

      if ((Self->Flags & MNF_SHOWKEYS) AND (Self->Items[i]->KeyString[0])) {
         if ((Self->Flags & MNF_SHOWIMAGES) AND (Self->ImageSize)) font->XCoord = x + Self->ImageSize + Self->ImageGap + Self->TextWidth + Self->KeyGap;
         else font->XCoord = x + Self->TextWidth + Self->KeyGap;

         font->YCoord = y + ((GetItemHeight(Self) - font->MaxHeight) / 2);
         SetField(font, FID_String, FT_STRING, Self->Items[i]->KeyString);
         font->AlignFlags |= ALIGN_RIGHT;
         font->AlignWidth = Self->KeyWidth;
         Action(AC_Draw, font, NULL);
         font->AlignFlags &= ~ALIGN_RIGHT;
      }

      /*** Draw menu item text ***/

      if (Self->Items[i]->Text) {
         if ((Self->Flags & MNF_SHOWIMAGES) AND (Self->ImageSize)) font->XCoord = x + Self->ImageSize + Self->ImageGap;
         else font->XCoord = x;

         font->YCoord = y + ((GetItemHeight(Self) - font->MaxHeight) / 2);
         SetField(font, FID_String, FT_STRING, Self->Items[i]->Text);

         Action(AC_Draw, font, NULL);
      }

      /*** Draw arrow for menu extensions ***/

      if (Self->Items[i]->Flags & MENUITEM_EXTENSION) {
         struct mtDrawRGBLine line;
         LONG ax, ay, aheight, awidth;

         awidth  = 5;
         aheight = 9;
         ax = Drawable->Width - Self->RightMargin - awidth - 4,
         ay = y + ((Self->Items[i]->Height - 9)/2);

         line.XCoord = ax + awidth - 1;
         line.YCoord = ay + (aheight>>1);
         line.XEnd   = ax;
         line.YEnd   = ay;
         if (Self->Highlight IS i) {
            line.Red    = Self->prvFontHighlight.Red;
            line.Green  = Self->prvFontHighlight.Green;
            line.Blue   = Self->prvFontHighlight.Blue;
         }
         else {
            line.Red    = Self->prvFontColour.Red;
            line.Green  = Self->prvFontColour.Green;
            line.Blue   = Self->prvFontColour.Blue;
         }

         if (aheight & 1) {
            while (line.YEnd < ay + aheight) {
               Action(MT_DrawRGBLine, Bitmap, &line);
               line.YEnd++;
            }
         }
         else {
            line.YCoord--;
            while (line.YEnd <= line.YCoord) {
               Action(MT_DrawRGBLine, Bitmap, &line);
               line.YEnd++;
            }

            line.YCoord++;
            while (line.YEnd < ay + aheight) {
               Action(MT_DrawRGBLine, Bitmap, &line);
               line.YEnd++;
            }
         }
      }

      y += Self->Items[i]->Height;
   }
}

/*****************************************************************************
** Internal: InitKeyboard()
*/

static void InitKeyboard(struct Menu *Self)
{
   OBJECTPTR keyboard;

   if ((Self->KeyboardID) AND (Self->KeyboardChannel IS FALSE)) {
      if (AccessObject(Self->KeyboardID, 4000, &keyboard) IS ERR_Okay) {
         if (SubscribeChannel(keyboard, Self) IS ERR_Okay) Self->KeyboardChannel = TRUE;
         ReleaseObject(keyboard);
      }
   }
}

/*****************************************************************************
** Internal: GetItemHeight()
*/

static WORD GetItemHeight(struct Menu *Self)
{
   WORD itemheight;
   itemheight = Self->Font->MaxHeight + Self->VSpacing;
   if ((Self->Flags & MNF_SHOWIMAGES) AND (itemheight < 16 + Self->VSpacing)) {
      itemheight = 16 + Self->VSpacing;
   }

   if (itemheight < Self->ItemHeight) itemheight = Self->ItemHeight;

   return(itemheight);
}

/*****************************************************************************
** Internal: SetKey()
*/

static ERROR SetKey(struct MenuItem *Item, STRING Value)
{
   LONG i;

   Item->Key = 0;
   if ((Value) AND (*Value)) {
      for (i=0; i < K_LIST_END; i++) {
         if (StrCompare(Value, glKeymapTable[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            Item->Key = i;
            return(ERR_Okay);
         }
      }
   }
   return(ERR_Search);
}

/*****************************************************************************
** Internal: SetQualifier()
*/

static ERROR SetQualifier(struct MenuItem *Item, STRING Value)
{
   LONG i;

   Item->QualifierFlags = 0;
   for (i=0; QualifierTable[i].Code != NULL; i++) {
      if (StrCompare(Value, QualifierTable[i].Name, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         Item->QualifierFlags |= QualifierTable[i].Code;
         return(ERR_Okay);
      }
   }
   return(ERR_Search);
}

/*****************************************************************************
** Internal: AddBreak()
*/

static UBYTE glMenuBreak[] = "<line points=\"(5,0) (!7,0) B (5,0) (!7,0)\" y0=\"={y}+2\" y1=\"={y}+2\" y2=\"={y}+3\" y3=\"={y}+3\" &col0=\"[gl_colours.shadow]\" &col2=\"[gl_colours.highlight]\"/>";

ERROR AddBreak(struct Menu *Self, struct MenuItem *Item)
{
   UBYTE buffer[30];
   struct Script *script;

   if (NewObject(ID_SCRIPT, NF_CHILD, &script, NULL) IS ERR_Okay) {
      if (Self->prvBreakScript) SetField(script, FID_String, FT_STRING, Self->prvBreakScript);
      else SetField(script, FID_String, FT_STRING, glMenuBreak);

      SetField(script, FID_Target, FT_LONG, &Self->DrawableID);

      IntToStr(Item->YCoord, buffer, sizeof(buffer));
      SetUnlistedField(script, "y", buffer);

      if (acInit(script) IS ERR_Okay) {
         acActivate(script);
      }

      acFree(script);
   }

   Item->Flags |= MENUITEM_BREAK;
   Item->Height = Self->BreakHeight;

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: AddXMLItem()
*/

static ERROR AddXMLItem(struct Menu *Self, OBJECTPTR XML, struct XMLTag *XMLTag)
{
   struct MenuItem item;
   struct mtGetXMLString getxml;
   STRING qualifier, key;
   LONG i, pos;

   if (!XMLTag) return(ERR_Args);

   if (!XMLTag->Attributes[0].Name) return(ERR_Okay);

   qualifier = NULL;
   key = NULL;
   ClearMemory(&item, sizeof(item));

   if (Self->TotalItems <= 0) item.YCoord = Self->TopMargin;
   else item.YCoord = Self->Items[Self->TotalItems-1]->YCoord + Self->Items[Self->TotalItems-1]->Height;

   if (StrCompare("if", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      if (IfStatementSatisfied(Self, XMLTag)) {
         XMLTag = XMLTag->Child;
         while (XMLTag) {
            AddXMLItem(Self, XML, XMLTag);
            XMLTag = XMLTag->Next;
         }
      }
      return(ERR_Okay);
   }
   else if (StrCompare("else", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      /*** Execute the contents of the <else> tag if the last <if> statement was not satisfied ***/
      if (glSatisfied IS FALSE) {
         XMLTag = XMLTag->Child;
         while (XMLTag) {
            AddXMLItem(Self, XML, XMLTag);
            XMLTag = XMLTag->Next;
         }
      }
      return(ERR_Okay);
   }
   else if (StrCompare("menu", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      for (i=1; i < XMLTag->TotalAttributes; i++) {
         if (StrCompare("icon", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (Self->Flags & MNF_SHOWIMAGES) LoadIcon(Self, XMLTag->Attributes[i].Value, &item.PictureID);
         }
         else if (StrCompare("text", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Text = Translate(Self, XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("location", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Location = StrClone(XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("categorise", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_CATEGORISE;
         }
         else if (StrCompare("sort", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_SORT;
         }
         else DPrintF("@Menu:","Unsupported menu attribute \"%s\".", XMLTag->Attributes[i].Name);
      }

      item.Flags |= MENUITEM_EXTENSION;
      item.Height = GetItemHeight(Self);

      if ((XML) AND (XMLTag->Child)) {
         struct mtGetXMLString getxml;
         getxml.Index = XMLTag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         item.ChildXML = getxml.Result;
      }

      if (AllocMemory(sizeof(item), MEM_DATA|MEM_NOCLEAR, &Self->Items[Self->TotalItems], NULL) IS ERR_Okay) {
         CopyMemory(&item, Self->Items[Self->TotalItems++], sizeof(item));
      }
   }
   else if (StrCompare("item", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      for (i=1; i < XMLTag->TotalAttributes; i++) {
         if (StrCompare("icon", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (Self->Flags & MNF_SHOWIMAGES) LoadIcon(Self, XMLTag->Attributes[i].Value, &item.PictureID);
         }
         else if (StrCompare("colour", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (ReadColour(XMLTag->Attributes[i].Value, &item.Colour.Red, &item.Colour.Green, &item.Colour.Blue) IS ERR_Okay) item.Flags |= MENUITEM_FONTCOLOUR;
         }
         else if (StrCompare("background", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (ReadColour(XMLTag->Attributes[i].Value, &item.Background.Red, &item.Background.Green, &item.Background.Blue) IS ERR_Okay) item.Flags |= MENUITEM_BACKGROUND;
         }
         else if (StrCompare("bold", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_FONTBOLD;
         }
         else if (StrCompare("italics", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_FONTITALICS;
         }
         else if (StrCompare("nokeyresponse", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_NOKEYRESPONSE;
         }
         else if (StrCompare("keyrepeat", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Flags |= MENUITEM_KEYREPEAT;
         }
         else if (StrCompare("text", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            item.Text = Translate(Self, XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("key", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (SetKey(&item, XMLTag->Attributes[i].Value) IS ERR_Okay) {
               key = XMLTag->Attributes[i].Value;
            }
         }
         else if (StrCompare("qualifier", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if (SetQualifier(&item, XMLTag->Attributes[i].Value) IS ERR_Okay) {
               qualifier = XMLTag->Attributes[i].Value;
            }
         }
      }

      if (key) {
         pos = 0;
         if (qualifier) {
            for (pos=0; (qualifier[pos]) AND (pos < LEN_KEYSTRING-2); pos++) item.KeyString[pos] = qualifier[pos];
            item.KeyString[pos++] = '+';
         }
         for (i=0; (key[i]) AND (pos < LEN_KEYSTRING-1); i++) item.KeyString[pos++] = key[i];
         item.KeyString[pos] = 0;
      }

      /*** Convert each XML tag into a command string ***/

      if (XMLTag->Child) {
         getxml.Index = XMLTag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         item.Commands = getxml.Result;
      }

      item.Height = GetItemHeight(Self);

      if ((XML) AND (XMLTag->Child)) {
         struct mtGetXMLString getxml;
         getxml.Index = XMLTag->Child->Index;
         getxml.Flags = XMF_INCLUDESIBLINGS;
         Action(MT_GetXMLString, XML, &getxml);
         item.ChildXML = getxml.Result;
      }

      if (AllocMemory(sizeof(item), MEM_DATA|MEM_NOCLEAR, &Self->Items[Self->TotalItems], NULL) IS ERR_Okay) {
         CopyMemory(&item, Self->Items[Self->TotalItems++], sizeof(item));
      }
   }
   else if (StrCompare("break", XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      AddBreak(Self, &item);

      if (AllocMemory(sizeof(item), MEM_DATA|MEM_NOCLEAR, &Self->Items[Self->TotalItems], NULL) IS ERR_Okay) {
         CopyMemory(&item, Self->Items[Self->TotalItems++], sizeof(item));
      }
   }
   else {
      DPrintF("@Menu:","Unsupported tag <%s>.", XMLTag->Attributes[0].Name);
      return(ERR_Okay);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CalculateMenuSize()
** Short:    Calculates the size of the menu's drawable based on the available items.
*/

static ERROR CalculateMenuSize(struct Menu *Self)
{
   struct acResize resize;
   LONG i, strwidth;

   if (Self->TotalItems <= 0) {
      resize.Width  = 100;
      resize.Height = Self->TopMargin + Self->BottomMargin + GetItemHeight(Self);
      resize.Depth  = 0;
      ActionMsg(AC_Resize, Self->DrawableID, &resize);
      return(ERR_Okay);
   }

   /* Calculate the correct width of the menu according to the text
   ** lengths and then perform the resize.
   */

   if (Self->Width) resize.Width = Self->Width;
   else {
      Self->TextWidth = 0;
      Self->KeyWidth = 0;

      resize.Width = 0;
      resize.Depth = 0;
      for (i=0; i < Self->TotalItems; i++) {
         if (Self->Items[i]->Text) {
            SetField(Self->Font, FID_String, FT_STRING, Self->Items[i]->Text);
            GetField(Self->Font, FID_Width, FT_LONG, &strwidth);

            strwidth += 8; /* Add a reasonable pixel gap so that text doesn't go too near the right edge */

            if (strwidth > Self->TextWidth) Self->TextWidth = strwidth;

            if ((Self->Items[i]->KeyString[0]) AND (Self->Flags & MNF_SHOWKEYS)) {
               SetField(Self->Font, FID_String, FT_STRING, Self->Items[i]->KeyString);
               GetField(Self->Font, FID_Width, FT_LONG, &strwidth);
               if (strwidth > Self->KeyWidth) Self->KeyWidth = strwidth;
            }
         }
      }

      if (Self->KeyWidth > 0) resize.Width = Self->TextWidth + Self->KeyGap + Self->KeyWidth;
      else resize.Width = Self->TextWidth;

      if (Self->Flags & MNF_SHOWIMAGES) resize.Width += Self->ImageSize + Self->ImageGap;

      resize.Width += Self->LeftMargin + Self->RightMargin;

      /*** If extension menu items are present then add some extra space for arrow graphics ***/

      for (i=0; i < Self->TotalItems; i++) {
         if (Self->Items[i]->Flags & MENUITEM_EXTENSION) {
            resize.Width += Self->ExtensionGap;
            break;
         }
      }
   }

   resize.Height = Self->Items[Self->TotalItems-1]->YCoord + Self->Items[Self->TotalItems-1]->Height + Self->BottomMargin;
   ActionMsg(AC_Resize, Self->DrawableID, &resize);

   EnsureVisibility(Self);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: EnsureVisibility()
** Short:    Prevents the menu from dropping off the edge of the screen.
*/

static void EnsureVisibility(struct Menu *Self)
{
   OBJECTPTR container, drawable;
   LONG y, containerheight, menuheight;

   if ((Self->TargetID) AND (Self->DrawableID)) {
      if (AccessObject(Self->TargetID, 5000, &container) IS ERR_Okay) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(container, FID_Height, FT_LONG, &containerheight);
            GetFields(drawable, FID_YCoord|TLONG, &y,
                                FID_Height|TLONG, &menuheight,
                                TAGEND);

            if ((y + menuheight) > containerheight) {
               /*** The menu goes past the viewable area, so reverse its position ***/
               y -= menuheight;
               if (Self->ParentID) y += GetItemHeight(Self) + Self->VWhiteSpace;
               SetField(drawable, FID_YCoord, FT_LONG, &y);
            }

            ReleaseObject(drawable);
         }
         else ObjectError(Self, ERH_Function, ERR_AccessObject);

         ReleaseObject(container);
      }
      else ObjectError(Self, ERH_Function, ERR_AccessObject);
   }
}

/*****************************************************************************
** Internal: CreateMenu()
*/

static UBYTE glMenuBase[] = "<dml>\n\
  <*render name=\"menu{id}\" x=\"0\" y=\"0\" width=\"100\" height=\"100\" sticktofront hide>
    <box boxes=\"(r0,0,!0,!0) (r1,1,!1,!1)\" thickness=\"1\" highlight(0)=\"200,200,200\"\n\
      shadow(0)=\"#000000\" colour(1)=\"[gl_colours.colour]\" highlight(1)=\"[gl_colours.highlight]\"\n\
      shadow(1)=\"[gl_colours.shadow]\"/>\n\
  </render>\n\
</dml>";

static ERROR CreateMenu(struct Menu *Self)
{
   struct XMLTag **xmllist;
   struct XMLTag *tag;
   struct Menu *parent;
   struct MenuItem *item;
   struct ConfigEntry *entries;
   struct RenderInfo info;
   OBJECTPTR drawable, script, config, relative;
   STRING section, qualifier, key;
   LONG i, j, amtentries, classid, pos, xcoord, ycoord;

   /*** Build the menu base ***/

   if (NewObject(ID_SCRIPT, NF_CHILD, &script, NULL) != ERR_Okay) return(ERR_NewObject);

   if (Self->prvBaseScript) {
      SetField(script, FID_String, FT_STRING, Self->prvBaseScript);
   }
   else SetField(script, FID_String, FT_STRING, glMenuBase);

   SetField(script, FID_Target, FT_LONG, &Self->TargetID);

   if (Action(AC_Init, script, NULL) IS ERR_Okay) {
      if (Action(AC_Activate, script, NULL) IS ERR_Okay) {
         if (GetField(script, FID_PrimaryObject, FT_LONG, &Self->DrawableID) != ERR_Okay) {
            Action(AC_Free, script, NULL);
            DPrintF("@Menu:","No primary object was specified in the base menu script.");
            return(ERR_Failed);
         }
      }
      else {
         Action(AC_Free, script, NULL);
         return(ERR_Activate);
      }
   }
   else {
      Action(AC_Free, script, NULL);
      return(ERR_Init);
   }

   Action(AC_Free, script, NULL);

   if ((Self->DrawableID) AND (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay)) {
      /* If there is a parent menu, route the focus through to it.  If not, route
      ** the focus through to the object that our menu is relative to.
      */

      GetRenderInfo(Self->DrawableID, &info, sizeof(info));

      if ((Self->ParentID) AND (AccessObject(Self->ParentID, 5000, (OBJECTPTR *)&parent) IS ERR_Okay)) {
         SetField(drawable, FID_RouteFocus, FT_LONG, &parent->DrawableID);
         ReleaseObject(parent);
      }
      else if (Self->RelativeID) SetField(drawable, FID_RouteFocus, FT_LONG, &Self->RelativeID);

      /* Calculate the correct coordinates for our menu.  This may mean
      ** retrieving the absolute coordinates of the relative drawable
      ** and using them to offset the menu coordinates.
      */

      if (Self->RelativeID) {
         if (AccessObject(Self->RelativeID, 5000, &relative) IS ERR_Okay) {
            if (GetFields(relative, FID_AbsX|TLONG, &xcoord,
                                    FID_AbsY|TLONG, &ycoord,
                                    TAGEND) IS ERR_Okay) {
               SetFields(drawable, FID_XCoord|TLONG, (LONG)(Self->XCoord + xcoord),
                                   FID_YCoord|TLONG, (LONG)(Self->YCoord + ycoord),
                                   TAGEND);
            }
            ReleaseObject(relative);
         }
      }
      else SetFields(drawable, FID_XCoord|TLONG, (LONG)Self->XCoord,
                               FID_YCoord|TLONG, (LONG)Self->YCoord,
                               TAGEND);

      Self->prvBitsPerPixel = info.BitsPerPixel;

      /* If the fade-in feature has been enabled, set the drawable's opacity to zero */

      if (Self->Flags & MNF_FADEIN) {
         SetFields(drawable, FID_Opacity|TLONG, 0,
                             TAGEND);
      }

      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Function, ERR_AccessObject));

   /* Identify the type of data that has been set in the location field.  Config and
   ** XML files are supported, but the developer can also write a complete XML string
   ** to the Location field if desired.
   */

   if (Self->Location) {
      if (*Self->Location IS '<') classid = ID_XML;
      else IdentifyFile(Self->Location, &classid, NULL);
   }
   else classid = NULL;

   /*** If a configuration file has been specified, load it ***/

   if (classid IS ID_CONFIG) {
      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, Self->Location,
         TAGEND) IS ERR_Okay) {

         if (GetFields(config, FID_Entries|TPTR, &entries,
                               FID_AmtEntries|TLONG, &amtentries,
                               TAGEND) IS ERR_Okay) {

            for (i=0; i < amtentries; i++) {
               section = entries[i].Section;
               if (AllocMemory(sizeof(struct MenuItem), MEM_DATA, &item, NULL) IS ERR_Okay) {

                  if (Self->TotalItems <= 0) item->YCoord = Self->TopMargin;
                  else item->YCoord = Self->Items[Self->TotalItems-1]->YCoord + Self->Items[Self->TotalItems-1]->Height;
                  item->Height = GetItemHeight(Self);

                  qualifier = NULL;
                  key = NULL;

                  if (StrCompare("break", entries[i].Section, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     AddBreak(Self, item);
                     while (i < amtentries) {
                        if (StrCompare("break", entries[i+1].Section, 0, STR_MATCHLENGTH) IS ERR_Okay) i++;
                        else break;
                     }
                  }
                  else while (i < amtentries)  {
                     if (StrCompare("text", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        item->Text = Translate(Self, entries[i].Data);
                     }
                     else if (StrCompare("icon", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (Self->Flags & MNF_SHOWIMAGES) LoadIcon(Self, entries[i].Data, &item->PictureID);
                     }
                     else if (StrCompare("command", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        UBYTE buffer[1000];
                        StrFormat(buffer, sizeof(buffer), "<%s/>", entries[i].Data);
                        item->Commands = StrClone(buffer);
                     }
                     else if (StrCompare("key", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (SetKey(item, entries[i].Data) IS ERR_Okay) key = entries[i].Data;
                     }
                     else if (StrCompare("qualifier", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (SetQualifier(item, entries[i].Data) IS ERR_Okay) qualifier = entries[i].Data;
                     }
                     else if (StrCompare("nokeyresponse", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        item->Flags |= MENUITEM_NOKEYRESPONSE;
                     }
                     else if (StrCompare("keyrepeat", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        item->Flags |= MENUITEM_KEYREPEAT;
                     }

                     if (StrCompare(section, entries[i+1].Section, 0, STR_MATCHLENGTH) IS ERR_Okay) i++;
                     else break;
                  }

                  if (key) {
                     pos = 0;
                     if (qualifier) {
                        for (pos=0; (qualifier[pos]) AND (pos < LEN_KEYSTRING-2); pos++) item->KeyString[pos] = qualifier[pos];
                        item->KeyString[pos++] = '+';
                     }
                     for (j=0; (key[j]) AND (pos < LEN_KEYSTRING-1); j++) item->KeyString[pos++] = key[j];
                     item->KeyString[pos] = 0;
                  }

                  Self->Items[Self->TotalItems++] = item;
               }
               else return(ERR_AllocMemory);
            }
         }

         Action(AC_Free, config, NULL);
      }
      else return(ERR_CreateObject);
   }
   else if ((classid IS ID_XML) OR (classid IS ID_SCRIPT)) {
      /* Proces the source file.  If the source is a script, we will
      ** assume it is an XML compatible DML file.
      */

      if (Self->Location[0] IS '<') SetField(Self->prvXML, FID_Statement, FT_STRING, Self->Location);
      else SetField(Self->prvXML, FID_Location, FT_STRING, Self->Location);

      if ((GetField(Self->prvXML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {

         /*** Find the first <menu> tag ***/

         for (i=0; xmllist[i]; i++) {
            if (StrCompare("menu", xmllist[i]->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               if (Self->prvNode[0]) {
                  /* If a node is specified, we have to check that the menu name matches the
                  ** node name.  If it doesn't, we'll keep searching for a specific menu definition.
                  */
                  for (j=1; j < xmllist[i]->TotalAttributes; j++) {
                     if (StrCompare("name", xmllist[i]->Attributes[j].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        if (StrCompare(Self->prvNode, xmllist[i]->Attributes[j].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                           break;
                        }
                     }
                  }

                  if (j < xmllist[i]->TotalAttributes) break;
               }
               else break;
            }
         }

         if (!xmllist[i]) {
            DPrintF("@Menu:","No <menu> tag was found in file \"%s\".", Self->Location);
            return(ERR_BadData);
         }

         for (tag=xmllist[i]->Child; (tag) AND (Self->TotalItems < MAX_ITEMS); tag=tag->Next) {
            AddXMLItem(Self, Self->prvXML, tag);
         }
      }
      else {
         DPrintF("@Menu:","Failed to retrieve any data from the XML definition.");
         return(ERR_GetField);
      }
   }
   else if (classid) DPrintF("@Menu:","File \"%s\" belongs to unsupported class #%d.", Self->Location, classid);

   /*** Recalculate the menu size ***/

   CalculateMenuSize(Self);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ExecuteItem()
**
** Index - The index of the menu item to execute.
** Focus - If TRUE and the item is for a sub-menu, the sub-menu will be given
**         the focus and keyboard control.  This is provided solely for occasions
**         where the user is using the keyboard to manipulate the menu system.
*/

static ERROR ExecuteItem(struct Menu *Self, LONG Index, LONG Focus)
{
   struct Menu *menu;
   struct Render *drawable;
   OBJECTPTR context, script;
   LONG xcoord, ycoord, width;
   OBJECTID currentmenu, menuid;
   ERROR error;

   if ((Index < 0) OR (Index > Self->TotalItems)) return(ERR_Okay);

   DPrintF("~ExecuteItem()","[Menu:%d] Executing item \"%s\".", Self->Head.UniqueID, Self->Items[Index]->Text);

   currentmenu = Self->CurrentMenuID;
   Self->CurrentMenuID = NULL;

   if (!Self->Items) {
      DPrintF("@ExecuteItem:","There are no items in this menu.");
      StepBack();
      return(ERR_Okay);
   }

   if (!Self->Items[Index]) {
      DPrintF("@ExecuteItem:","Item %d has no associated functionality.", Index);
      StepBack();
      return(ERR_Okay);
   }

   if (Self->Items[Index]->Flags & MENUITEM_EXTENSION) {
      DPrintF("ExecuteItem:","Item is a menu extension.");

      /*** Hide any currently open sub-menu ***/

      if ((currentmenu) AND (currentmenu != Self->Items[Index]->MenuID)) {
         ActionMsg(AC_Hide, currentmenu, NULL);
      }

      if (AccessObject(Self->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
         GetFields(drawable, FID_XCoord|TLONG, &xcoord,
                             FID_Width|TLONG,  &width,
                             FID_YCoord|TLONG, &ycoord,
                             TAGEND);
         ReleaseObject(drawable);
      }
      else return(StepObjectError(Self, ERH_Function, ERR_AccessObject));

      /* Either set up the existing sub-menu or create a new one if it
      ** does not exist yet.
      */

      if ((menuid = Self->Items[Index]->MenuID)) {
         DPrintF("ExecuteItem:","Activating existing child menu #%d.", menuid);

         if (AccessObject(Self->Items[Index]->MenuID, 5000, (OBJECTPTR *)&menu) IS ERR_Okay) {
            /*** Hide any active sub menus ***/

            if (menu->CurrentMenuID) {
               ActionMsg(AC_Hide, menu->CurrentMenuID, NULL);
               menu->CurrentMenuID = NULL;
            }

            /*** Display the selected menu item, so long as it is not already selected ***/

            if (Focus) menu->Highlight = 0;

            if (AccessObject(menu->DrawableID, 5000, (OBJECTPTR *)&drawable) IS ERR_Okay) {
               if (!(drawable->Flags & RNF_VISIBLE)) {
                  SetFields(drawable, FID_XCoord|TLONG, xcoord + width - Self->RightMargin,
                                      FID_YCoord|TLONG, ycoord + Self->Items[Index]->YCoord,
                                      TAGEND);
                  EnsureVisibility(menu);
                  acMoveToFront(drawable);
                  acShow(drawable);
                  DelayMsg(AC_Focus, menu->DrawableID, NULL); /* Delay required to avoid recursive keypresses */
               }
               ReleaseObject(drawable);
            }
            else return(StepObjectError(Self, ERH_Function, ERR_AccessObject));

            Self->CurrentMenuID = Self->Items[Index]->MenuID;
            ReleaseObject(menu);
         }
         else Self->Items[Index]->MenuID = NULL; /* The child menu has been killed */
      }

      if (!Self->Items[Index]->MenuID) {
         if (NewObject(ID_MENU, NF_CHILD, &menu, &menuid) IS ERR_Okay) {
            SetName(menu, "submenu");
            menu->TargetID = Self->TargetID;
            menu->ParentID = Self->Head.UniqueID;
            menu->RootID   = Self->RootID;
            menu->Flags    = Self->Flags;
            menu->XCoord   = xcoord + width - Self->RightMargin;
            menu->YCoord   = ycoord + Self->Items[Index]->YCoord;
            menu->VSpacing = Self->VSpacing;
            menu->VWhiteSpace      = Self->VWhiteSpace;
            menu->ParentIndex      = Index;
            menu->KeyboardID       = Self->KeyboardID;
            menu->KeyMonitorID     = Self->KeyMonitorID;
            menu->LeftMargin       = Self->LeftMargin;
            menu->TopMargin        = Self->TopMargin;
            menu->BottomMargin     = Self->BottomMargin;
            menu->RightMargin      = Self->RightMargin;
            menu->ImageGap         = Self->ImageGap;
            menu->KeyGap           = Self->KeyGap;
            menu->ExtensionGap     = Self->ExtensionGap;
            menu->HighlightLM      = Self->HighlightLM;
            menu->HighlightRM      = Self->HighlightRM;
            menu->ItemHeight       = Self->ItemHeight;
            menu->BreakHeight      = Self->BreakHeight;
            menu->AutoExpand       = Self->AutoExpand;
            menu->Duration         = Self->Duration;
            menu->ImageSize        = Self->ImageSize;
            menu->prvFontColour    = Self->prvFontColour;
            menu->prvFontHighlight = Self->prvFontHighlight;
            menu->prvHighlight     = Self->prvHighlight;
            menu->prvHighlightBorder = Self->prvHighlightBorder;

            /*** The base script defines the base graphics for the menu ***/

            if (Self->prvBaseScript) {
               context = SetContext(menu);
                  menu->prvBaseScript = StrClone(Self->prvBaseScript);
               SetContext(context);
            }

            /*** The break script defines the graphical look of menu-breaks ***/

            if (Self->prvBreakScript) {
               context = SetContext(menu);
                  menu->prvBreakScript = StrClone(Self->prvBreakScript);
               SetContext(context);
            }

            /* If the menu refers to a configuration file that needs to be categorised,
            ** we need to load the config file and turn it into an XML-Menu definition
            ** file.
            */

            if (Self->Items[Index]->Flags & MENUITEM_CATEGORISE) {
               if ((error = GenerateMenuFile(Self, menu, Index)) != ERR_Okay) {
                  acFree(menu);
                  ReleaseObject(menu);
                  return(StepObjectError(Self, ERH_Function, error));
               }
            }
            else {
               if (Self->Items[Index]->Location) {
                  SetField(menu, FID_Location, FT_STRING, Self->Items[Index]->Location);
               }

               if (acInit(menu) != ERR_Okay) {
                  acFree(menu);
                  ReleaseObject(menu);
                  return(StepObjectError(Self, ERH_Function, ERR_Init));
               }
            }

            /* If there are child tags in our menu that we need to associate
            ** with this sub-menu, add them into the sub-menu item list.  This
            ** is where each <item> tag is added.
            */

            if (Self->Items[Index]->ChildXML) {
               struct XMLTag *tag;
               struct XMLTag **xmllist;

               SetField(Self->prvXML, FID_Statement, FT_STRING, Self->Items[Index]->ChildXML);

               context = SetContext(menu); /* Ensure that any allocations are against the sub-menu, not us */

                  if ((GetField(Self->prvXML, FID_Tags, FT_POINTER, &xmllist) IS ERR_Okay) AND (xmllist)) {
                     for (tag=xmllist[0]; tag; tag=tag->Next) AddXMLItem(menu, Self->prvXML, tag);
                     CalculateMenuSize(menu);
                  }

               SetContext(context);
            }

            /*** Put the menu on display ***/

            if (Focus) {
               menu->Highlight = 0;
               acFocusID(menu->Head.UniqueID);
            }

            if (ActionMsg(AC_Show, menu->DrawableID, NULL) IS ERR_Okay) {
               Self->CurrentMenuID = menuid;
            }

            Self->Items[Index]->MenuID = menuid;

            ReleaseObject(menu);
         }
         else return(StepObjectError(Self, ERH_Function, ERR_NewObject));
      }

      /*** If we are focussing on the sub-menu, kill our current focus ***/

      if (Focus) {
         HighlightItem(Self, -1);
      }
   }
   else {
      /* Instantly hide the root menu drawable (no fading).  We also
      ** switch the focus to the object that we are relative to.
      */

      if (Self->RootID IS Self->Head.UniqueID) {
         ActionMsg(AC_Hide, Self->DrawableID, NULL);
         if (Self->RelativeID) ActionMsg(AC_Focus, Self->RelativeID, NULL);
      }
      else if (AccessObject(Self->RootID, 4000, (OBJECTPTR *)&menu) IS ERR_Okay) {
         ActionMsg(AC_Hide, menu->DrawableID, NULL);
         if (menu->RelativeID) ActionMsg(AC_Focus, menu->RelativeID, NULL);
         ReleaseObject(menu);
      }

      if (Self->SelectionObjectID) {
         OBJECTPTR object;
         if (AccessObject(Self->SelectionObjectID, 5000, &object) IS ERR_Okay) {
            if (Self->SelectionField[0]) {
               SetFieldVariable(object, Self->SelectionField, Self->Items[Index]->Text);
            }
            else SetField(object, FID_String, FT_STRING, Self->Items[Index]->Text);
            ReleaseObject(object);
         }
      }

      if (Self->Items[Index]->Commands) {
         if (NewObject(ID_SCRIPT, NF_CHILD, &script, NULL) IS ERR_Okay) {
            STRING value;
            WORD i;

            SetField(script, FID_Statement, FT_STRING, Self->Items[Index]->Commands);

            /*** Pass our unlisted arguments on to the script ***/

            for (i=0; i < Self->TotalArgs; i++) {
               if (Self->LocalArgs[i]) {
                  value = Self->LocalArgs[i];
                  while (*value) value++;
                  value++;
                  SetUnlistedField(script, Self->LocalArgs[i], value);
               }
            }

            if (Action(AC_Init, script, NULL) IS ERR_Okay) {
               Action(AC_Activate, script, NULL);
            }

            Action(AC_Free, script, NULL);
         }
      }
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: GenerateMenuFile()
**
** This function turns configuration files into menu files.  The menu is
** sorted and organised according to the Category item in each section.
** Multiple categories are allowed to organise the menu structure into
** sub-trees, e.g. "Development/SDK/Documentation"
*/

#define SIZE_MENU_BUFFER 2000

static void WriteMenuItems(struct Menu *, OBJECTPTR, OBJECTPTR, STRING, STRING *, WORD *, struct ConfigEntry *entries, LONG amtentries);

static ERROR GenerateMenuFile(struct Menu *Self, struct Menu *Menu, LONG Index)
{
   OBJECTPTR config, file;
   struct ConfigEntry *entries;
   STRING *list;
   LONG amtentries, total, j, i, pos;
   WORD index;
   UBYTE buffer[SIZE_MENU_BUFFER];
   ERROR error;

   DPrintF("GenerateMenuFile()","Index: %d", Index);

   error = ERR_Failed;

   if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
      FID_Location|TSTRING, Self->Items[Index]->Location,
      TAGEND) != ERR_Okay) return(ERR_CreateObject);

   /* Sort the configuration file immediately after loading.  Note
   ** that sorting occurs on the configuration section names.
   */

   if ((Self->Flags & MNF_SORT) OR (Self->Items[Index]->Flags & MENUITEM_SORT)) {
      struct mtSortConfig sort;
      sort.Item = "Text";
      sort.Descending = FALSE;
      Action(MT_SortConfig, config, &sort);
   }

   if (GetFields(config, FID_Entries|TPTR,     &entries,
                         FID_AmtEntries|TLONG, &amtentries,
                         TAGEND) != ERR_Okay) {
      acFree(config);
      return(ERR_GetField);
   }

   /* Gather all category fields in the config file into a sequential
   ** string list that we can send to StrBuildArray()
   */

   pos   = 0;
   total = 0;
   for (i=0; i < amtentries; i++) {
      if (StrCompare("category", entries[i].Item,  0, STR_MATCHLENGTH) IS ERR_Okay) {
         for (j=0; (entries[i].Data[j]) AND (pos < sizeof(buffer)-1); j++) buffer[pos++] = entries[i].Data[j];
         buffer[pos++] = 0;
         total++;
      }
   }

   if ((list = StrBuildArray(buffer, pos, total, SBF_SORT|SBF_NODUPLICATES))) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, "temp:menu.xml",
         FID_Flags|TLONG,      FL_NEW|FL_WRITE,
         TAGEND) IS ERR_Okay) {

         WriteFile(file, "<?xml version=\"1.0\"?>\n\n");
         WriteFile(file, "<menu>\n");

         index = 0;
         while (list[index]) {
            WriteMenuItems(Self, config, file, buffer, list, &index, entries, amtentries);
         }

         WriteFile(file, "</menu>\n");

         SetField(Menu, FID_Location, FT_STRING, "temp:menu.xml");

         if (acInit(Menu) != ERR_Okay) {
            acFree(Menu);
            ReleaseObject(Menu);
            return(ERR_Init);
         }

         Action(MT_DeleteFile, file, NULL);
         acFree(file);

         error = ERR_Okay;
      }
      else error = ERR_CreateObject;

      FreeMemory(list);
   }
   else error = ERR_BadData;

   return(error);
}

static void WriteMenuItems(struct Menu *Self, OBJECTPTR config, OBJECTPTR file, STRING Buffer,
               STRING *List, WORD *Index, struct ConfigEntry *entries, LONG amtentries)
{
   struct mtReadConfig read;
   STRING category, path, str;
   LONG i, section;

   category = List[*Index];
   for (i=0; category[i]; i++);
   while ((i > 0) AND (category[i-1] != '/')) i--;

   StrFormat(Buffer, SIZE_MENU_BUFFER, "  <menu text=\"%s\" icon=\"icons:folders/programfolder\">\n", category+i);
   WriteFile(file, Buffer);

   /*** Test the next category in the list.  If it is a sub-category, recurse into it ***/

   path = List[Index[0]];
   while (List[Index[0]+1]) {
      str = List[Index[0]+1];
      for (i=0; (path[i]) AND (path[i] IS str[i]); i++);
      if ((!path[i]) AND (str[i] IS '/')) {
         /*** We've found a sub-category ***/
         Index[0] = Index[0] + 1;
         WriteMenuItems(Self, config, file, Buffer, List, Index, entries, amtentries);
         Index[0] = Index[0] - 1;
      }
      else break;
   }

   /*** Write out all items in the current category ***/

   section = 0;
   for (i=0; i < amtentries; i++) {
      if (StrCompare(entries[i].Section, entries[section].Section, 0, STR_MATCHLENGTH) != ERR_Okay) {
         section = i;
      }

      if (StrCompare("category", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (StrCompare(category, entries[i].Data, 0, STR_MATCHLENGTH) IS ERR_Okay) {

            WriteFile(file, "    <item");

            read.Section = entries[i].Section;
            read.Item    = "Icon";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               StrFormat(Buffer, SIZE_MENU_BUFFER, " icon=\"%s\"", read.Data);
               WriteFile(file, Buffer);
            }

            read.Section = entries[i].Section;
            read.Item    = "Text";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               StrFormat(Buffer, SIZE_MENU_BUFFER, " text=\"%s\"", read.Data);
               WriteFile(file, Buffer);
            }

            WriteFile(file, ">\n");

            read.Section = entries[i].Section;
            read.Item    = "Command";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               StrFormat(Buffer, SIZE_MENU_BUFFER, "      <%s/>\n", read.Data);
               WriteFile(file, Buffer);
            }

            WriteFile(file, "    </item>\n");
         }
      }
   }

   WriteFile(file, "  </menu>\n\n");

   /*** Increment the current list position before returning ***/

   Index[0] = Index[0] + 1;
}

/*****************************************************************************
** Internal: LoadIcon()
**
** The location of the icon should be supplied in the format:
**
**    icons:category/name
**
** Any size references will be stripped from the location.
*/

static ERROR LoadIcon(struct Menu *Self, STRING Location, OBJECTID *PictureID)
{
   struct Picture *picture;
   UBYTE buffer[300];
   WORD i;

   /* Load the icon graphic for this menu.  Failure is not
   ** terminal if the image cannot be loaded - the image will simply
   ** not appear in the menu bar.
   */

   StrCopy(Location, buffer, sizeof(buffer));

   /*** Find the end of the location ***/

   for (i=0; buffer[i]; i++);

   /*** Strip out any existing size references ***/

   if (buffer[i-1] IS ')') {
      while ((i > 0) AND (buffer[i] != '(')) i--;
      if (buffer[i] IS '(') buffer[i] = 0;
   }

   /*** Add our own predefined size reference ***/

   buffer[i++] = '(';
   buffer[i++] = '1';
   buffer[i++] = '6';
   buffer[i++] = ')';
   buffer[i++] = 0;

   if (NewObject(ID_PICTURE, NULL, &picture, PictureID) IS ERR_Okay) {
      SetFields(picture, FID_Location|TSTRING, buffer, TAGEND);
      picture->Bitmap->BitsPerPixel  = Self->prvBitsPerPixel;
      picture->Flags |= PCF_32BITALPHA|PCF_CACHE;

      if (Action(AC_Init, picture, NULL) IS ERR_Okay) {

      }
      else Action(AC_Free, picture, NULL);

      ReleaseObject(picture);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: WriteFile()
*/

ERROR WriteFile(OBJECTPTR File, STRING String)
{
   struct acWrite write;
   write.Buffer = String;
   for (write.Length=0; String[write.Length]; write.Length++);
   return(Action(AC_Write, File, &write));
}

/****************************************************************************/

static ERROR HighlightItem(struct Menu *Self, LONG Index)
{
   struct acDraw draw;

   if (Index > Self->TotalItems) Index = Self->TotalItems - 1;
   if (Index < -1) Index = -1;

   if ((Index != -1) AND (Self->Items[Index]->Flags & MENUITEM_BREAK)) return(ERR_Okay);

   if (Index IS Self->Highlight) return(ERR_Okay);

   /*** Redraw the previously highlighted item ***/

   if (Self->Highlight != -1) {
      draw.XCoord = 0;
      draw.YCoord = Self->Items[Self->Highlight]->YCoord;
      draw.Width  = 10000;
      draw.Height = Self->Items[Self->Highlight]->Height;

      Self->Highlight = -1;
      ActionMsg(AC_Draw, Self->DrawableID, &draw);
   }

   Self->Highlight = Index;

   if (Index IS -1) return(ERR_Okay);

   /*** Draw the newly highlighted area ***/

   draw.XCoord = 0;
   draw.YCoord = Self->Items[Index]->YCoord;
   draw.Width  = 10000;
   draw.Height = Self->Items[Index]->Height;
   ActionMsg(AC_Draw, Self->DrawableID, &draw);

   return(ERR_Okay);
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING Str)
{
   WORD Value;
   if (!*Str) return(0);
   if ((*Str >= '0') AND (*Str <= '9')) Value = (*Str - '0') * 16;
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value = ((*Str - 'A')+10) * 16;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value = ((*Str - 'a')+10) * 16;
   else Value = 0;
   Str++;
   if ((*Str >= '0') AND (*Str <= '9')) Value += (*Str - '0');
   else if ((*Str >= 'A') AND (*Str <= 'F')) Value += (*Str - 'A') + 10;
   else if ((*Str >= 'a') AND (*Str <= 'f')) Value += (*Str - 'a') + 10;
   return(Value);
}

/****************************************************************************/

static ERROR ReadColour(STRING Colour, LONG *Red, LONG *Green, LONG *Blue)
{
   if (*Colour IS '#') {
      Colour++;
      *Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      *Blue  = ReadByte(Colour);
   }
   else {
      *Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      *Blue = StrToInt(Colour);
   }
   return(ERR_Okay);
}
