#ifndef  MODULES_SNAP_H
#define  MODULES_SNAP_H

/*
**   snap.h
**
**   (C) Copyright 2002-2003 Rocklyte Systems.
**       All Rights Reserved.
*/

#ifndef  MAIN_H
#include <pandora/main.h>
#endif

#include <pandora/graphics/bitmap.h>

#define MODVERSION_SNAP ((FLOAT)1.0)

#define MAX_ADDRESSMAPS 60

typedef struct SNAPOptimal {
   void (*PutPixel)(N_int32, N_int32);
   void (*SetPixelColour)(GA_color);
   void (*DrawLine)(N_int32, N_int32, N_int32, N_int32, N_int32);
   N_int32 (*DrawLineClip)(N_int32 x1,N_int32 y1,N_int32 x2,N_int32 y2,N_int32 drawLast,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
   void (*SetLineColour)(GA_color);
   void (*SetLineClipColour)(GA_color);
   void (*SetMonoColour)(GA_color);
   void (*DrawRect)(N_int32, N_int32, N_int32, N_int32);
   void (*DrawRectLin)(N_int32 dstOfs,N_int32 dstPitch,N_int32 left,N_int32 top,N_int32 width,N_int32 height,GA_color color,N_int32 mix);
   void (*SetRectColour)(GA_color);
   void (*SrcTransBlt)(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
   void (*BitBlt)(N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
   void (*BitBltLin)(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix);
   void (*SrcTransBltSys)(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent,N_int32 flipY);
   void (*SrcTransBltLin)(N_int32 srcOfs,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,GA_color transparent);
   void (*BitBltSys)(void *srcAddr,N_int32 srcPitch,N_int32 srcLeft,N_int32 srcTop,N_int32 width,N_int32 height,N_int32 dstLeft,N_int32 dstTop,N_int32 mix,N_int32 flipY);
   void (*ClipMonoImageSys)(N_int32 x,N_int32 y,N_int32 width,N_int32 height,N_int32 byteWidth,N_uint8 *image,N_int32 transparent,N_int32 clipLeft,N_int32 clipTop,N_int32 clipRight,N_int32 clipBottom);
} GA_optimalFuncs;

struct SNAPGlobals {
   APTR Device;                   /* Device address */
   GA_modeInfo      VideoMode;    /* Current video mode information */
   GA_initFuncs     Init;
   GA_bufferFuncs   Buffer;
   GA_cursorFuncs   Cursor;
   GA_driverFuncs   Driver;
   GA_2DStateFuncs  gsState;
   GA_2DRenderFuncs gsDraw;
   GA_2DStateFuncs  gaState;
   GA_2DRenderFuncs gaDraw;
   GA_optimalFuncs  Optimal;
   GA_DPMSFuncs     gsDPMS;
   struct {
      APTR Address;             /* Fixed address to map from memory, NULL if random allocation */
      LONG Length;              /* Size */
      APTR MapAddress;          /* Process address that received the mapping */
      LONG ID;
   } AddressMaps[MAX_ADDRESSMAPS];
   UBYTE HWCursor;
   UBYTE CursorVisible;
   WORD  CursorX;
   WORD  CursorY;
   OBJECTID LastBitmapID;
   LONG     MemoryKey;
   MEMORYID CursorMID;
};

typedef enum {
   SNAP_INIT = 1,
   SNAP_DRIVER,
   SNAP_BUFFER,
   SWSNAP_2DSTATE,
   SWSNAP_2DRENDER,
   HWSNAP_2DSTATE,
   HWSNAP_2DRENDER,
   HWSNAP_CURSOR,
   SNAP_EXPORTS,
   SNAP_OPTIMAL
} SNAPFunctions;

#ifndef LOCAL_SNAP

struct SNAPBase {
   ERROR (*OpenSNAP)(GA_devCtx **);
   ERROR (*GetFunctions)(LONG, APTR *);
   ERROR (*OpenVideoMode)(LONG, GA_modeInfo *, LONG *, LONG *, LONG *, LONG *, FLOAT);
   ERROR (*CloseVideoMode)(LONG);
   ERROR (*AllocVideo)(LONG, LONG, GA_buf **);
   ERROR (*LockSNAP)(void);
   ERROR (*GetVideoMode)(GA_modeInfo **);
   ERROR (*SetDisplayState)(LONG);
   void  (*ShowCursor)(LONG);
   void  (*UnlockSNAP)(void);
   ERROR (*SetMonoCursor)(GA_monoCursor *);
   void  (*FreeVideo)(GA_buf *);
   ERROR (*SetCursor)(APTR, LONG, LONG, LONG, LONG, LONG, LONG);
   LONG  (*GetCursorDepth)(void);
};

#define sciAllocVideo(a,b,c)            (SNAPBase->AllocVideo(a,b,c))
#define sciCloseVideoMode(a)            (SNAPBase->CloseVideoMode(a))
#define sciFreeVideo(a)                 (SNAPBase->FreeVideo(a))
#define sciGetCursorDepth()             (SNAPBase->GetCursorDepth())
#define sciGetFunctions(a,b)            (SNAPBase->GetFunctions(a,b))
#define sciGetVideoMode(a)              (SNAPBase->GetVideoMode(a))
#define sciLockSNAP()                   (SNAPBase->LockSNAP())
#define sciOpenSNAP(a)                  (SNAPBase->OpenSNAP(a))
#define sciOpenVideoMode(a,b,c,d,e,f,g) (SNAPBase->OpenVideoMode(a,b,c,d,e,f,g))
#define sciSetCursor(a,b,c,d,e,f,g)     (SNAPBase->SetCursor(a,b,c,d,e,f,g))
#define sciSetDisplayState(a)           (SNAPBase->SetDisplayState(a))
#define sciSetMonoCursor(a)             (SNAPBase->SetMonoCursor(a))
#define sciShowCursor(a)                (SNAPBase->ShowCursor(a))
#define sciUnlockSNAP()                 (SNAPBase->UnlockSNAP())

#else

static ERROR sciAllocVideo(LONG, LONG, GA_buf **);
static ERROR sciCloseVideoMode(LONG);
static void sciFreeVideo(GA_buf *);
static LONG sciGetCursorDepth(void);
static ERROR sciGetFunctions(LONG, APTR *);
static ERROR sciGetVideoMode(GA_modeInfo **);
static ERROR sciLockSNAP(void);
static ERROR sciOpenSNAP(GA_devCtx **);
static ERROR sciOpenVideoMode(LONG, GA_modeInfo *, LONG *, LONG *, LONG *, LONG *, FLOAT);
static ERROR sciSetCursor(struct Bitmap *, LONG, LONG, LONG, LONG, LONG, LONG);
static ERROR sciSetDisplayState(LONG);
static ERROR sciSetMonoCursor(GA_monoCursor *);
static void sciShowCursor(LONG);
static void sciUnlockSNAP(void);

#endif /* LOCAL_SNAP */

#endif /* MODULES_SNAP_H */
